#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass

from entity_read import sql
from entity_read.entity import Entity
from entity_read.sql.atoms import Selectable
from .expression import Expression

key_to_function_factory = {
    "lt": sql.inter.FunctionLessThan,
    "le": sql.inter.FunctionLessEqual,
    "eq": sql.inter.FunctionEqual,
    "ne": sql.inter.FunctionNonEqual,
    "ge": sql.inter.FunctionGreaterEqual,
    "gt": sql.inter.FunctionGreater,
    "in": sql.inter.FunctionIn,
    "not_in": sql.inter.FunctionNotIn,
    "or": sql.inter.FunctionOr,
    "and": sql.inter.FunctionAnd,
    "not": sql.inter.FunctionNot,
    "like": lambda args: sql.inter.FunctionStringContains(
        args=(sql.inter.FunctionCastToStr(args=args[:1]), args[1]), ignore_case=False),
    "ilike": lambda args: sql.inter.FunctionStringContains(
        args=(sql.inter.FunctionCastToStr(args=args[:1]), args[1]), ignore_case=True),
    "is_null": sql.inter.FunctionIsNull,
    "is_not_null": sql.inter.FunctionIsNotNull,
}


@dataclass(frozen=True, kw_only=True, repr=False)
class Condition(Expression):
    key: str
    args: list[Expression]

    def eval(self, entity_type: type[Entity], variables: dict[str, Selectable]) -> sql.atoms.function.Function:
        if (function_factory := key_to_function_factory.get(self.key)) is None:
            raise KeyError(f"Not found condition {self.key!r}")
        # noinspection PyArgumentList
        return function_factory(args=[arg.eval(entity_type, variables) for arg in self.args])

    def shortcut(self) -> str:
        return f"cond.{self.key}({','.join([arg.shortcut() for arg in self.args])})"



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass

from entity_read import sql
from entity_read.entity import Entity
from entity_read.sql.atoms import Selectable
from .expression import Expression


@dataclass(frozen=True, kw_only=True, repr=False)
class Variable(Expression):
    key: str

    def eval(self, entity_type: type[Entity], variables: dict[str, Selectable]) -> sql.atoms.column.Selectable:
        if (var := variables.get(self.key)) is None:
            raise KeyError(f"Not found variable {self.key}, present: {list(variables.keys())}")
        if isinstance(var, sql.lower.LowerSelector):
            return sql.atoms.column.Column(key=self.key, type_=list)
        if isinstance(var, sql.atoms.Selectable):
            return var
        raise TypeError(f"Unexpected type: {type(var)}")

    def shortcut(self) -> str:
        return f"var.{self.key}"



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import typing

from entity_read import sql
from entity_read.entity import Entity

from .expression import Expression
from .order import Order
from .relation import Relation

if typing.TYPE_CHECKING:
    from .subquery import SubQuery


def eval_to_selector(
        entity_type: type[Entity], local_column_to_remote: dict[str, str],
        attrs: list[Expression], vars_: dict[str, Expression], filters: list[Expression], searches: list[Expression],
        orders: list[Order], limit: int | None, offset: int | None,
        my_step: str | None = None, rename: str | None = None, is_scalar: bool = False
) -> sql.lower.LowerSelector:
    from .subquery import SubQuery
    variables = {}
    for key, val in vars_.items():
        variables[key] = val.eval(entity_type, variables)

    where = [attr.eval(entity_type, variables) for attr in filters]
    if searches:
        where.append(sql.inter.FunctionOr(args=tuple(attr.eval(entity_type, variables) for attr in searches)))

    result_name_to_selectable = {}
    step_to_sub_selector = {}
    solo_result = None
    if isinstance(attrs, Expression):
        node = attrs.eval(entity_type, variables)
        if isinstance(node, sql.atoms.Selectable | sql.lower.LowerSelector):
            solo_result = node
        else:
            raise TypeError("got unexpected type")
    else:
        for attr in attrs:
            node = attr.eval(entity_type, variables)
            if isinstance(node, sql.atoms.Selectable):
                result_name_to_selectable[node.get_name()] = node
            elif isinstance(node, sql.lower.LowerSelector):
                assert isinstance(attr, SubQuery)
                if isinstance(attr.over, Relation):
                    step_to_sub_selector[attr.over.key] = node
                else:
                    step_to_sub_selector[f'ref_{attr.over.key}'] = node
            else:
                raise TypeError("got unexpected type")

    for key, var in variables.items():
        if isinstance(var, sql.lower.LowerSelector):
            assert key not in step_to_sub_selector
            step_to_sub_selector[key] = var
            var.is_hidden = True

    result = sql.lower.LowerSelector(
        table_name=entity_type.get_table_name(), schema_name=entity_type.get_table_schema_name(),
        result_name_to_selectable=result_name_to_selectable,
        where=where, is_hidden=False, order=[order.eval(entity_type, variables) for order in orders],
        limit=limit, offset=offset, my_step=my_step, rename=rename, is_scalar=is_scalar,
        local_column_to_remote=local_column_to_remote, step_to_sub_selector=step_to_sub_selector,
        solo_result=solo_result
    )
    return result



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass

from entity_read.sql.atoms import Selectable

from entity_read import sql
from entity_read.entity import Entity
from .expression import Expression
from .query_token import QueryToken


@dataclass(frozen=True, kw_only=True, repr=False)
class Order(QueryToken):
    body: Expression
    asc: bool = True
    nulls_last: bool | None = None

    def eval(self, entity_type: type[Entity], variables: dict[str, Selectable]) -> sql.atoms.order.Order:
        return sql.atoms.order.Order(
            value=self.body.eval(entity_type, variables), asc=self.asc, nulls_last=self.nulls_last)

    def shortcut(self) -> str:
        nulls_last = f",nulls_last={self.nulls_last}" if self.nulls_last is not None else ''
        return f"sql.order({self.body.shortcut()},asc={self.asc}{nulls_last})"



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
import abc
import typing
from dataclasses import dataclass

from entity_read import sql
from entity_read.entity import Entity
from entity_read.sql.atoms import Selectable
from .query_token import QueryToken


@dataclass(frozen=True, kw_only=True, repr=False)
class Expression(QueryToken, abc.ABC):

    @abc.abstractmethod
    def eval(self, entity_type: type[Entity], variables: dict[str, Selectable]) -> sql.atoms.Selectable:
        pass


def as_(self, key: str):
    from .alias import Alias
    return Alias(body=self, key=key)


setattr(Expression, 'as', as_)


def make_condition_generator(condition_key: str) -> typing.Callable[[Expression, typing.Any], Expression]:
    def condition_generator(self: Expression, *args: typing.Any):
        from .condition import Condition
        from .literal import Literal
        right_args = [a if isinstance(a, Expression) else Literal(body=a) for a in args]
        return Condition(
            args=[self] + right_args,
            key=condition_key,
        )
    return condition_generator


def make_and_set_condition_method(target: type[Expression], condition_key: str) -> None:
    setattr(target, condition_key, make_condition_generator(condition_key))


for key in ['eq', 'ne', 'gt', 'lt', 'ge', 'le', 'in', 'is_null', 'is_not_null', 'ilike', 'like']:
    make_and_set_condition_method(Expression, key)



#  Copyright (C) 2023
#  ABM, Moscow
#
#  UNPUBLISHED PROPRIETARY MATERIAL.
#  ALL RIGHTS RESERVED.
#
#  Authors: Mike Orlov <m.orlov@abm-jsc.ru>
from dataclasses import dataclass, field

from entity_read import sql
from entity_read.entity import Entity
from .expression import Expression
from .query_token import QueryToken
from .order import Order
from .subquery import prepare_tokens
from ._eval_to_selector import eval_to_selector


@dataclass(frozen=True, kw_only=True, repr=False)
class Query(QueryToken):
    # entity: str
    attrs: list[Expression] | Expression = field(default_factory=list)
    vars: dict[str, Expression] = field(default_factory=dict)
    filters: list[Expression] = field(default_factory=list)
    searches: list[Expression] = field(default_factory=list)
    orders: list[Order] = field(default_factory=list)
    limit: int | None = None
    offset: int | None = None
    is_scalar: bool | None = None

    def __post_init__(self):
        if self.is_scalar is None:
            object.__setattr__(self, 'is_scalar', not isinstance(self.attrs, list))

    def shortcut(self) -> str:
        # parts = [self.entity]
        parts = []
        parts += prepare_tokens('attrs', self.attrs)
        parts += [f'vars={",".join([f"{k}={v.shortcut()}" for k, v in self.vars.items()])}'] if self.vars else []
        parts += prepare_tokens('filters', self.filters)
        parts += prepare_tokens('searches', self.searches)
        parts += prepare_tokens('orders', self.orders)
        parts += [f'limit={self.limit}'] if self.limit else []
        parts += [f'offset={self.offset}'] if self.offset else []
        parts += [f'is_scalar={self.is_scalar}'] if self.is_scalar else []
        # parts += [f'attrs={[attr.shortcut() for attr in self.attrs]}'] if self.attrs else []
        # parts += [f'vars={[f"{key}={val.shortcut()}" for key, val in self.vars.items()]}'] if self.vars else []
        # parts += [f'filters={[attr.shortcut() for attr in self.filters]}'] if self.filters else []
        # parts += [f'searches={[attr.shortcut() for attr in self.searches]}'] if self.searches else []
        # parts += [f'orders={[attr.shortcut() for attr in self.orders]}'] if self.orders else []
        return f'Query({",".join(parts)})'

    def eval(self, entity_type: type[Entity]) -> sql.lower.LowerSelector:
        return eval_to_selector(
            entity_type=entity_type, local_column_to_remote={},
            attrs=self.attrs, vars_=self.vars, filters=self.filters, searches=self.searches,
            orders=self.orders, limit=self.limit, offset=self.offset, is_scalar=self.is_scalar
        )

        # variables = {}
        # for key, val in self.vars.items():
        #     variables[key] = val.eval(entity_type, variables)
        #
        # where = [attr.eval(entity_type, variables) for attr in self.filters]
        # if self.searches:
        #     where.append(sql.inter.FunctionOr(args=tuple(attr.eval(entity_type, variables) for attr in self.searches)))
        #
        # result_name_to_selectable = {}
        # step_to_sub_selector = {}
        # for attr in self.attrs:
        #     node = attr.eval(entity_type, variables)
        #     if isinstance(node, sql.atoms.Selectable):
        #         result_name_to_selectable[node.get_name()] = node
        #     elif isinstance(node, sql.lower.LowerSelector):
        #         assert isinstance(attr, SubQuery)
        #         step_to_sub_selector[attr.over.eval(entity_type, variables).get_name()] = node
        #     else:
        #         raise TypeError("goy unexpected type")
        #
        # result = sql.lower.LowerSelector(
        #     table_name=entity_type.get_table_name(), result_name_to_selectable=result_name_to_selectable,
        #     where=where, is_hidden=False, order=[attr.eval(entity_type, variables) for attr in self.order],
        #     limit=self.limit, offset=self.offset,
        #     local_column_to_remote={}, step_to_sub_selector=step_to_sub_selector
        # )
        # return result
